#ifndef __Mt_Luts_H__
#define __Mt_Luts_H__

#include "../../../common/base/filter.h"
#include "../../../../common/parser/parser.h"

#include "../functions.h"

namespace Filtering { namespace MaskTools { namespace Filters { namespace Lut { namespace Spatial {

typedef void(Processor)(Byte *pDst, int nDstPitch, const Byte *pSrc, int nSrcPitch, const Byte lut[65536], const int *pCoordinates, int nCoordinates, int nWidth, int nHeight);

extern Processor *deviater_c;
extern Processor *averager_c;
extern Processor *minimizer_c;
extern Processor *maximizer_c;
extern Processor *ranger_c;
extern Processor *medianizer_c;
extern Processor *medianizer2_c;
extern Processor *medianizer4_c;
extern Processor *medianizer6_c;

class Filter : public MaskTools::Filter<InPlaceFilter>
{
   Byte luts[3][65536];

   int *pCoordinates;
   int nCoordinates;

   ProcessorList<Processor> processors;

   void FillCoordinates(const String &coordinates)
   {
      std::list<Parser::Symbol> coeffs = Parser::getDefaultParser().parse( coordinates, " (),;." ).getExpression();
      nCoordinates = coeffs.size();
      pCoordinates = new int[nCoordinates];
      int i = 0;

      while ( !coeffs.empty() )
      {
         pCoordinates[i++] = int( coeffs.front().getValue(0, 0, 0) );
         coeffs.pop_front();
      }
   }

protected:
   virtual void process(int n, const Plane<Byte> &dst, int nPlane)
   {
      processors.best_processor( constraints[nPlane] )( dst, dst.pitch(), frames[0].plane(nPlane), frames[0].plane(nPlane).pitch(), luts[nPlane], pCoordinates, nCoordinates, dst.width(), dst.height() );
   }

public:
   Filter(const Parameters &parameters) : MaskTools::Filter<InPlaceFilter>( parameters )
   {
      static const char *expr_strs[] = { "yExpr", "uExpr", "vExpr" };

      Parser::Parser parser = Parser::getDefaultParser().addSymbol(Parser::Symbol::X).addSymbol(Parser::Symbol::Y);

      /* compute the luts */
      for ( int i = 0; i < 3; i++ )
      {
         if ( parameters[expr_strs[i]].is_defined() ) 
            parser.parse(parameters[expr_strs[i]].toString(), " ");
         else
            parser.parse(parameters["expr"].toString(), " ");

         Parser::Context ctx(parser.getExpression());

         if ( !ctx.check() )
         {
            error = "invalid expression in the lut";
            return;
         }

         for ( int x = 0; x < 256; x++ )
            for ( int y = 0; y < 256; y++ )
               luts[i][(x<<8)+y] = ctx.compute_byte(x, y);
      }

      /* get the pixels list */
      FillCoordinates( parameters["pixels"].toString() );

      /* choose the mode */
      String mode = parameters["mode"].toString();
      if ( mode == "average" || mode == "avg" )
         processors.push_back( Filtering::Processor<Processor>( averager_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else if ( mode == "standard deviation" || mode == "std" )
         processors.push_back( Filtering::Processor<Processor>( deviater_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else if ( mode == "minimum" || mode == "min" )
         processors.push_back( Filtering::Processor<Processor>( minimizer_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else if ( mode == "maximum" || mode == "max" )
         processors.push_back( Filtering::Processor<Processor>( maximizer_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else if ( mode == "range" )
         processors.push_back( Filtering::Processor<Processor>( ranger_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else if ( mode == "median" || mode == "med" )
         processors.push_back( Filtering::Processor<Processor>( medianizer4_c, Constraint( CPU_NONE, 1, 1, 1, 1 ), 0 ) );
      else
      {
         error = "unknown mode";
         return;
      }
   }

   InputConfiguration &input_configuration() const { return InPlaceTwoFrame(); }

   static Signature filter_signature()
   {
      Signature signature = "mt_luts";

      signature.add( Parameter( TYPE_CLIP, "" ) );
      signature.add( Parameter( TYPE_CLIP, "" ) );
      signature.add( Parameter( String( "average" ), "mode" ) );
      signature.add( Parameter( String( "" ), "pixels" ) );
      signature.add( Parameter( String( "y" ), "expr" ) );
      signature.add( Parameter( String( "y" ), "yExpr" ) );
      signature.add( Parameter( String( "y" ), "uExpr" ) );
      signature.add( Parameter( String( "y" ), "vExpr" ) );

      return add_defaults( signature );
   }
};

} } } } }

#endif